from asyncio import Event
from contextlib import asynccontextmanager
from typing import (
    AsyncIterable,
    Awaitable,
    Callable,
    Generic,
    Tuple,
    Type,
    TypeVar,
)

from fifolock import FifoLock

from kilroy_server_py_utils.locks import Read, Write
from kilroy_server_py_utils.observable import (
    FetchableObservable,
    Observable,
    ReadOnlyObservableWrapper,
    ReadableObservable,
)

T = TypeVar("T")
LoadableType = TypeVar("LoadableType", bound="Loadable")


class EventBasedObservableWrapper(FetchableObservable[T], Generic[T]):
    _observable: Observable[T]
    _event: Event

    def __init__(self, observable: Observable, event: Event) -> None:
        self._observable = observable
        self._event = event

    async def fetch(self) -> T:
        await self._event.wait()
        return await self._observable.fetch()

    async def subscribe(self) -> AsyncIterable[T]:
        async for value in self._observable.subscribe():
            yield value


class ValueWrapper(Generic[T]):
    _value: T

    def __init__(self, getter: Callable[[], Awaitable[T]]) -> None:
        super().__init__()
        self._getter = getter

    @property
    def value(self) -> T:
        return self._value

    @property
    def has_value(self) -> bool:
        return hasattr(self, "_value")

    async def get(self) -> T:
        return await self._getter()

    async def set(self, value: T) -> None:
        self._value = value


class Loadable(Generic[T]):
    _value: Observable[T]
    _ready: Observable[bool]
    _event: Event
    _lock: FifoLock

    def __init__(
        self,
        value: Observable[T],
        ready: Observable[bool],
        event: Event,
        lock: FifoLock,
    ) -> None:
        self._value = value
        self._ready = ready
        self._event = event
        self._lock = lock
        self._event.set()

    @classmethod
    async def build(cls: Type[LoadableType], **kwargs) -> LoadableType:
        return cls(
            value=await Observable.build(),
            ready=await Observable.build(False),
            event=Event(),
            lock=FifoLock(),
            **kwargs,
        )

    @property
    def value(self) -> FetchableObservable[T]:
        return EventBasedObservableWrapper(self._value, self._event)

    @property
    def ready(self) -> ReadableObservable[bool]:
        return ReadOnlyObservableWrapper(self._ready)

    @asynccontextmanager
    async def read_lock(self) -> T:
        async with self._lock(Read):
            yield await self._value.fetch()

    @asynccontextmanager
    async def write_lock(self) -> T:
        async with self._lock(Write):
            yield await self._value.fetch()

    @asynccontextmanager
    async def load(
        self,
    ) -> Tuple[Callable[[], Awaitable[T]], Callable[[], Awaitable[None]]]:
        async with self._lock(Write):
            self._event.clear()
            ready = await self._ready.get()
            await self._ready.set(False)
            try:
                wrapper = ValueWrapper(self._value.get)
                yield wrapper.get, wrapper.set
                if wrapper.has_value:
                    await self._value.set(wrapper.value)
                    ready = True
            finally:
                await self._ready.set(ready)
                self._event.set()



import inspect
from typing import Any, Dict, Generic, Set, Type, TypeVar

from kilroy_server_py_utils.loadable import Loadable
from kilroy_server_py_utils.observable import (
    Observable,
    ReadOnlyObservableWrapper,
    ReadableObservable,
)
from kilroy_server_py_utils.parameters.base import Parameter
from kilroy_server_py_utils.schema import JSONSchema
from kilroy_server_py_utils.utils import classproperty, get_generic_args

T = TypeVar("T")
StateType = TypeVar("StateType")
ConfigurationType = TypeVar("ConfigurationType", bound="Configuration")
ConfigurableType = TypeVar("ConfigurableType", bound="Configurable")


class Configuration(Generic[StateType]):
    _parameters: Set[Parameter]
    _state: Loadable[StateType]
    _json: Observable[Dict[str, Any]]

    def __init__(
        self,
        parameters: Set[Parameter],
        state: Loadable[StateType],
        json: Observable[Dict[str, Any]],
    ) -> None:
        self._parameters = parameters
        self._state = state
        self._json = json

    @classmethod
    async def build(
        cls: Type[ConfigurationType], parameters: Set[Parameter], **kwargs
    ) -> ConfigurationType:
        return cls(
            parameters=parameters,
            state=await Loadable.build(),
            json=await Observable.build(),
            **kwargs,
        )

    @property
    def state(self) -> Loadable[StateType]:
        return self._state

    @property
    def json(self) -> ReadableObservable[Dict[str, Any]]:
        return ReadOnlyObservableWrapper(self._json)

    async def _build_json(self) -> Dict[str, Any]:
        state = await self._state.value.fetch()
        return {
            parameter.name: await parameter.get(state)
            for parameter in self._parameters
        }

    async def init(self, state: StateType) -> None:
        async with self._state.load() as (_, setter):
            await setter(state)
        await self._json.set(await self._build_json())

    async def set(self, config: Dict[str, Any]) -> Dict[str, Any]:
        params_map = {
            parameter.name: parameter for parameter in self._parameters
        }

        async with self._state.load() as (getter, setter):
            state = await getter()
            undos = []

            try:
                for name, value in config.items():
                    parameter = params_map.get(name, None)
                    if parameter is not None:
                        undo = await parameter.set(state, value)
                        undos.append(undo)
                await setter(state)
            except Exception as e:
                for undo in reversed(undos):
                    try:
                        await undo()
                    except Exception:
                        pass
                raise e

        config = await self._build_json()
        await self._json.set(config)
        return config


class Configurable(Generic[StateType]):
    _config: Configuration[StateType]

    def __init__(self, config: Configuration[StateType], **kwargs) -> None:
        self._config = config
        self._kwargs = kwargs

    @classmethod
    async def build(
        cls: Type[ConfigurableType], *args, **kwargs
    ) -> ConfigurableType:
        return cls(
            config=await Configuration.build(cls.parameters),
            *args,
            **kwargs,
        )

    @property
    def config(self) -> Configuration[StateType]:
        return self._config

    @property
    def state(self) -> Loadable[StateType]:
        return self._config.state

    async def build_default_state(self) -> StateType:
        # noinspection PyUnresolvedReferences
        return get_generic_args(self, Configurable)[0](**self._kwargs)

    async def init(self) -> None:
        state = await self.build_default_state()
        await self._config.init(state)

    async def cleanup(self) -> None:
        pass

    @classproperty
    def parameters(cls) -> Set[Parameter]:
        return {
            inner_cls()
            for inner_cls in cls.__dict__.values()
            if inspect.isclass(inner_cls) and issubclass(inner_cls, Parameter)
        }

    @classproperty
    def schema_name(cls) -> str:
        return f"{cls.__name__} config schema"

    @classproperty
    def schema(cls) -> JSONSchema:
        return JSONSchema(
            title=cls.schema_name,
            type="object",
            properties=cls.properties_schema,
        )

    @classproperty
    def properties_schema(cls) -> Dict[str, Any]:
        return {
            parameter.name: parameter.schema for parameter in cls.parameters
        }



import inspect
from asyncio import AbstractEventLoop, get_running_loop
from base64 import urlsafe_b64decode, urlsafe_b64encode
from concurrent.futures import Executor
from functools import partial
from typing import (
    Callable,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    _GenericAlias,
    get_args,
)

from humps import camelize, kebabize

T = TypeVar("T")


def get_generic_args(ref: Union[T, Type[T]], cls: Type[T]) -> Tuple:
    bases = ref.__orig_bases__
    generic_base = next(
        base
        for base in bases
        if isinstance(base, _GenericAlias) and issubclass(base.__origin__, cls)
    )
    return get_args(generic_base)


def can_call_with_arg(f: Callable, arg_name: str) -> bool:
    try:
        inspect.signature(f).bind_partial(**{arg_name: None})
    except TypeError:
        return False

    return True


class classproperty:
    def __init__(self, method=None):
        self.fget = method

    def __get__(self, instance, cls=None):
        return self.fget(cls)

    def getter(self, method):
        self.fget = method
        return self


def base64_encode(value: bytes) -> str:
    return urlsafe_b64encode(value).decode("ascii")


def base64_decode(value: str) -> bytes:
    return urlsafe_b64decode(value.encode("ascii"))


async def noop(*args, **kwargs) -> None:
    pass


async def background(
    f: Callable[..., T],
    *args,
    loop: Optional[AbstractEventLoop] = None,
    executor: Optional[Executor] = None,
    **kwargs,
) -> T:
    loop = loop or get_running_loop()
    f = partial(f, *args, **kwargs)
    return await loop.run_in_executor(executor, f)


def normalize(name: str) -> str:
    name = name.lower() if name.isupper() else name
    return camelize(kebabize(name))



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['kilroy_server_py_utils',
 'kilroy_server_py_utils.parameters',
 'kilroy_server_py_utils.resources']

package_data = \
{'': ['*']}

install_requires = \
['fifolock>=0.0,<0.1',
 'jsonschema>=4.7,<5.0',
 'pydantic>=1.9,<2.0',
 'pyhumps>=3.7,<4.0']

extras_require = \
{'dev': ['pytest>=7.0,<8.0'], 'test': ['pytest>=7.0,<8.0']}

setup_kwargs = {
    'name': 'kilroy-server-py-utils',
    'version': '0.1.3',
    'description': 'utilities for kilroy servers 🔧',
    'long_description': '<h1 align="center">kilroy-server-py-utils</h1>\n\n<div align="center">\n\nutilities for kilroy servers in Python 🔧\n\n[![Tests](https://github.com/kilroybot/kilroy-server-py-utils/actions/workflows/test-multiplatform.yml/badge.svg)](https://github.com/kilroybot/kilroy-server-py-utils/actions/workflows/test-multiplatform.yml)\n[![Docs](https://github.com/kilroybot/kilroy-server-py-utils/actions/workflows/docs.yml/badge.svg)](https://github.com/kilroybot/kilroy-server-py-utils/actions/workflows/docs.yml)\n\n</div>\n\n---\n\nTODO\n\n## Installing\n\nUsing `pip`:\n\n```sh\npip install kilroy-server-py-utils\n```\n',
    'author': 'kilroy',
    'author_email': 'kilroymail@pm.me',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/kilroybot/kilroy-server-py-utils',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
